; #########################################################################
;
;                             qw2a
;
;Procedure for converting a 64-bit signed integer from memory to ASCII
;
;Limitation: The st(6) and st(7) registers on the FPU will be trashed
;
;Return: EAX!=0 and null-terminated ASCII string in memory
;        EAX =0 if FPU unusable and null string in memory
;
;Usage:  invoke qw2a,lpQW,lpBuf
;
;   where:  
;   lpQW  = pointer to the location of the qword in memory
;   lpBuf = pointer to a memory buffer for the null-terminated ASCII
;           string (a 21-byte buffer is sufficient)
;
;                       by Raymond Filiatreault
;                              May 2004
;                            MASM  syntax
;
; #########################################################################
;
;      .386                   ; minimum processor needed for 32 bit
;      .model flat, stdcall   ; FLAT memory model & STDCALL calling
;      option casemap :none   ; set code to case sensitive
;
; #########################################################################

.code

qw2a proc USES edi lpQW:DWORD, lpBuf:DWORD

LOCAL power9      :DWORD
LOCAL float0      :DWORD
LOCAL lowdw       :DWORD
LOCAL highdw      :DWORD
LOCAL highdw2     :DWORD
LOCAL tempqw      :QWORD
LOCAL oldcw       :WORD
LOCAL roundcw     :WORD
LOCAL trunccw     :WORD

      fclex             ;clear exception flags
      ffree st(7)
      ffree st(6)       ;2 FPU registers required

      mov   eax,lpQW    ;retrieve address of qword
      fild  qword ptr[eax] ;load the qword on the FPU
      fstsw ax          ;to check for exception flags
      fwait
      shr   al,1
      jnc   @F          ;continue if loading accepted by FPU

      xor   eax,eax
      mov   edi,lpBuf
      mov   [edi],al
      ret

   @@:
      fstcw oldcw          ;get control word
      fwait
      mov   ax,oldcw
      and   ah,0f3h     ;set rounding control to rounding
      mov   roundcw,ax
      or    ah,0ch      ;set rounding control to truncating
      mov   trunccw,ax

      fst   float0      ;copy to memory as REAL4 to retrieve sign later
      fabs              ;insure it is positive
      mov   power9,1000000000

      fldcw roundcw     ;load rounding control word
      fidiv power9      ;get the value of the 10^9 digits
      fld   st          ;copy result in 2nd FPU register
      fldcw trunccw     ;load truncating control word
      fistp tempqw      ;store the truncated integer value >10^9 digits
      fldcw roundcw     ;load rounding control word
      
      fild  tempqw      ;reload it
      fsubp st(1),st    ;keep only the fractional portion (lower 9 digits)
      fimul power9      ;restore the integer value of those digits
      fistp lowdw       ;store it

      fild  tempqw      ;load the >10^9 digits
      fidiv power9      ;get the value of the digit higher than 10^18 
      fldcw trunccw     ;load truncating control word
      fist  highdw2     ;store the value of a 19th digit
      fisub highdw2     ;keep only the fractional part (2nd 9 digits)
      fldcw roundcw     ;load rounding control word
      fimul power9      ;restore the integer value of those 9 digits
      fistp highdw

      fldcw oldcw       ;restore the original control word

      mov   eax,lowdw   ;lower 9 digits now in EAX

;check for a 0 value

      or    eax,eax
      jnz   @F
      cmp   highdw,0
      jnz   @F
      cmp   highdw2,0
      jnz   @F

      mov   edi,lpBuf
      mov   eax,"0"
      mov   [edi],ax
      ret

   @@:
      mov   edi,9       ;counter for lower 9 digits
      mov   ecx,10
      pushd 0           ;for later use as terminating 0
   @@:
      xor   edx,edx     ;clear for division
      div   ecx         ;remainder (less than 10) -> EDX (effectively DL)
      add   dl,"0"      ;convert to ascii
      push  edx         ;save each digit on stack
      dec   edi         ;bump down count of digits
      jz    @F
      or    eax,eax
      jnz   @B
      cmp   highdw,0
      jnz   @B
      cmp   highdw2,0
      jnz   @B
      jmp   wq2ax

   @@:
      mov   edi,9       ;counter for next 9 digits
      mov   eax,highdw
      or    eax,eax
      jnz   @F
      cmp   highdw2,0
      jz    wq2ax
   @@:
      xor   edx,edx     ;clear for division
      div   ecx         ;remainder (less than 10) -> EDX (effectively DL)
      add   dl,"0"      ;convert to ascii
      push  edx         ;save each digit on stack
      dec   edi         ;bump down count of digits
      jz    @F
      or    eax,eax
      jnz   @B
      cmp   highdw2,0
      jnz   @B
      jmp   wq2ax

   @@:
      mov   eax,highdw2
      or    eax,eax
      jz    wq2ax
   @@:
      xor   edx,edx     ;clear for division
      div   ecx         ;remainder (less than 10) -> EDX (effectively DL)
      add   dl,"0"      ;convert to ascii
      push  edx         ;save each digit on stack
      or    eax,eax
      jnz   @B

wq2ax:
   @@:
      mov   edi,lpBuf
      mov   eax,float0  ;get the copy of the float
      or    eax,eax     ;check for its sign
      jns   @F
      mov   al,"-"
      stosb
   @@:
      pop   eax         ;retrieve each ascii character from the stack
      stosb
      or    al,al
      jnz   @B          ;continue until terminating 0 is retrieved
      or    al,1
      ret

qw2a endp

;end
